/***********************************************************************************************************************
 *
 * blueBill Resources - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~resources-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.factsheet.wikipedia.photos;

import java.io.IOException;
import javax.annotation.Nonnull;
import it.tidalwave.bluebill.taxonomy.Taxon;
import it.tidalwave.bluebill.taxonomy.Taxon.Builder;
import it.tidalwave.bluebill.taxonomy.birds.col.CatalogueOfLifeImporter;
import it.tidalwave.openrdf.elmo.impl.ElmoEntityFactory;
import it.tidalwave.semantic.EntityFactory;
import it.tidalwave.semantic.Wrapper;
import it.tidalwave.util.Initializer;
import it.tidalwave.xml.XPathProvider;
import it.tidalwave.xml.XmlParser;
import it.tidalwave.xml.dom4j.Dom4jXmlParser;
import it.tidalwave.xml.jaxen.JaxenXPathProvider;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.Proxy;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.xml.namespace.QName;
import org.apache.commons.io.IOUtils;
import org.openrdf.concepts.dc.DcResource;
import org.openrdf.concepts.foaf.FoafResource;
import org.openrdf.concepts.foaf.Image;
import org.openrdf.concepts.skos.core.Concept;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.jaxen.XPath;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
public class WikipediaPhotosImporter extends CatalogueOfLifeImporter
  {
    public static final String DCMI_TYPE_STILL_IMAGE = "http://purl.org/dc/dcmitype/StillImage";
    public static final String LICENSE_GNU_FDL = "http://www.gnu.org/copyleft/fdl.html";
    
    @Nonnull
    private final transient XmlParser xmlParser;

    @Nonnull
    private final transient XPathProvider xPathProvider;

    private transient Proxy proxy = Proxy.NO_PROXY;

    private static final String XPATH_DIV = "/html:html/html:body/html:div/html:div/html:table/html:tr/html:td/html:div/html:div/html:div";
    private static final String XPATH_A_HREF = "html:a/@href";
    private static final String XPATH_A_IMG_SRC = "html:a/html:img/@src";
    private static final String XPATH_2 = "//html:div[@id='file']/html:a[1]/@href";
    private static final String XPATH_3 = "/html:html/html:head/html:link[@rel='copyright']/@href";

    protected final transient EntityFactory<Description, ElmoDescription> stillImageFactory =
            new ElmoEntityFactory<Description, ElmoDescription>(Description.class,
                                                    ElmoDescription.class,
                                                    Image.class,
                                                    DcResource.class);

    protected final transient EntityFactory<Organization, ElmoOrganization> organizationFactory =
            new ElmoEntityFactory<Organization, ElmoOrganization>(Organization.class,
                                                              ElmoOrganization.class,
                                                              Concept.class,
                                                              DcResource.class,
                                                              org.openrdf.concepts.foaf.Organization.class,
                                                              FoafResource.class);

    protected final transient EntityFactory<Person, ElmoPerson> authorFactory =
            new ElmoEntityFactory<Person, ElmoPerson>(Person.class,
                                                              ElmoPerson.class,
                                                              Concept.class,
                                                              DcResource.class,
                                                              org.openrdf.concepts.foaf.Person.class);

    protected transient Organization wikiPedia;

    private transient final Map<String, Person> map = new HashMap<String, Person>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected final transient EntityResolver entityResolver = new EntityResolver()
      {
        @Nonnull
        public InputSource resolveEntity (final @Nonnull String publicId, final @Nonnull String systemId)
          throws SAXException, IOException
          {
            if ("http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd".equals(systemId))
              {
                return new InputSource(getClass().getResourceAsStream("xhtml1-transitional.dtd"));
              }

            return null;
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public WikipediaPhotosImporter()
      throws IOException
      {
        this(new Dom4jXmlParser(), new JaxenXPathProvider());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ WikipediaPhotosImporter (final @Nonnull XmlParser xmlParser,
                                           final @Nonnull XPathProvider xPathProvider)
      throws IOException
      {
        this.xmlParser = xmlParser;
        this.xPathProvider = xPathProvider;
        this.xmlParser.setEntityResolver(entityResolver);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    protected void initialize()
      throws IOException
      {
        super.initialize();
        wikiPedia = organizationFactory.create(new QName("http://wikipedia.org/#organization"));
        wikiPedia.getLookup().lookup(Concept.class).setRdfsLabel("Wikipedia");
//        wikiPedia.getLookup().lookup(Organization.class).getFoafHomepages().add("http://wikipedia.org");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected Builder process (final @Nonnull String path,
                               final @Nonnull String scientificName,
                               @Nonnull Builder builder)
      throws Exception
      {
        if (builder.getRank() != Taxon.Rank.SPECIES)
          {
            throw new IgnoreException();
          }

        final URL url = new URL("http://commons.wikimedia.org/wiki/" + scientificName.replace(' ', '_'));
        final InputStream is = url.openStream();
        final List<URI> uris = processPage(is);
        is.close();

        builder = builder.withInitializer(new Initializer<Taxon>()
          {
            @Nonnull
            public Taxon initialize (final @Nonnull Taxon taxon)
              {
                for (final URI uri : uris)
                  {
                    try
                      {
                      System.err.printf("Processing: %s\n", uri);
                        final InputStream is2 = uri.toURL().openStream();
                        final Description media = processPage2(is2);
                        is2.close();
                        media.getLookup().lookup(Image.class).getFoafDepicts().add(Wrapper.unwrap(taxon));
                      }
                    catch (Exception e)
                      {
                        e.printStackTrace();
                      }
                  }
                
                return taxon;
              }
          });

        return builder;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected List<URI> processPage (final @Nonnull InputStream is)
      throws Exception
      {
        final Object document = xmlParser.parse(is);
        final XPath exprDiv     = xPathProvider.createXPath(document, XPATH_DIV);
        final XPath exprAHref   = xPathProvider.createXPath(document, XPATH_A_HREF);
        final XPath exprAImgSrc = xPathProvider.createXPath(document, XPATH_A_IMG_SRC);

        final List<Object> trs = exprDiv.selectNodes(document);
        final List<URI> result = new ArrayList<URI>();

        for (int i = 0; i < trs.size(); i++)
          {
            final Object tr = trs.get(i);
            final String wikiPage = "http://commons.wikimedia.org" + exprAHref.stringValueOf(tr);
            final String thumbnail = exprAImgSrc.stringValueOf(tr);
              System.err.println(wikiPage + "  " + thumbnail);

              result.add(new URI(wikiPage));
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected Description processPage2 (final @Nonnull InputStream is)
      throws Exception
      {
        final byte[] bytes = IOUtils.toByteArray(is);
        final String string = new String(bytes, "UTF-8");
        final Object document = xmlParser.parse(new ByteArrayInputStream(bytes));
        final XPath exprImageUrl   = xPathProvider.createXPath(document, XPATH_2);
        final XPath exprLicenseUrl = xPathProvider.createXPath(document, XPATH_3);

        final String imageUrl = exprImageUrl.stringValueOf(document);
        final Description media = stillImageFactory.create(new QName(imageUrl));
        final DcResource dcResource = media.getLookup().lookup(DcResource.class); // FIXME: as()
        dcResource.getDcTypes().add(DCMI_TYPE_STILL_IMAGE);
        dcResource.getDcFormats().add("image/jpeg");
        dcResource.getDcPublishers().add(Wrapper.unwrap(wikiPedia));
        dcResource.getDcIdentifiers().add(imageUrl);

//          System.err.println("IMAGE " + imageUrl);
        final String licenseUrl = exprLicenseUrl.stringValueOf(document);
//          System.err.println("LICENSE " + licenseURL);
        dcResource.getDcRights().add(licenseUrl);

        int h = string.indexOf("Creator:");
        
        if (h >= 0)
          {
            h += "Creator:".length();
            int j = string.substring(h).indexOf("</");
            String author = string.substring(h, h + j).trim();
            System.err.println("AUTHOR: " + author);
            final String authorUri = "http://wikipedia.org/contributor/" + author;
            dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, author)));
          }
        
        else
          {
            int i = string.indexOf("Author:");

            if (i >= 0)
              {
                i += "Author:".length();
                int j = string.substring(i).indexOf("</p>");
                String author = string.substring(i, i + j).trim();

                if (author.contains("href"))
                  {
                    author = author.replace("</b> ", "");
                    int i1 = author.indexOf('"');
                    int i2 = author.substring(i1 + 1).indexOf('"') + i1;
                    final String authorUri = "http://commons.wikimedia.org" + author.substring(i1 + 1, i2 + 1);
                    int j1 = author.indexOf('>');
                    int j2 = author.substring(j1 + 1).indexOf('<') + j1;
                    final String authorName = author.substring(j1 + 1, j2 + 1);
                    dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, authorName)));
                    System.err.println("AUTHOR: " + authorName);
                  }
                else
                  {
                    System.err.println("AUTHOR: " + author);
                    final String authorUri = "http://wikipedia.org/contributor/" + author;
                    dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, author)));
                  }
              }
            
            else
              {
                int j = string.indexOf("Author</th>");

                if (j >= 0)
                  {
                    final String author = string.substring(j + 12);
                    int i1 = author.indexOf("href=\"") + 5;
                    int i2 = author.substring(i1 + 1).indexOf('"') + i1;
                    final String authorUri = "http://commons.wikimedia.org" + author.substring(i1 + 1, i2 + 1);
                    int j1 = author.indexOf("\">") + 1;
                    int j2 = author.substring(j1 + 1).indexOf('<') + j1;
                    final String authorName = author.substring(j1 + 1, j2 + 1);
                    dcResource.getDcCreators().add(Wrapper.unwrap(findOrCreateAuthor(authorUri, authorName)));
                    System.err.println("AUTHOR: " + authorName);
                  }
              }
          }

        return media;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private Person findOrCreateAuthor (final @Nonnull String uri, final @Nonnull String name)
      {
        Person foafEntity = map.get(uri);

        if (foafEntity == null)
          {
            foafEntity = authorFactory.create(new QName(uri));
            map.put(uri, foafEntity);
//            foafEntity.getLookup().lookup(Concept.class).setRdfsLabel(name);
            foafEntity.getLookup().lookup(org.openrdf.concepts.foaf.Person.class).getFoafGivennames().add(name);
          }

        return foafEntity;
      }
  }
